#!/usr/bin/env python3

"""  od2ri
  (absorption) optical depth to radiation intensity --- Schwarzschild solver

  usage:
  od2ri   [options]  opticalDepthFile

  -h              help
  -c    char      comment character(s) used in input,output file (default '#')
  -o    string    output file for saving of radiances (if not given: write to StdOut)

 --BoA  float     bottom-of-atmosphere altitude [km]  (read opt.depth only for levels above)
 --ToA  float     top-of-atmosphere altitude [km]     (read opt.depth only for levels below)
                  NOTE:  no interpolation, i.e. integration starts/stops at the next level above/below BoA/ToA
  -C              flag indicating that input optical depth is cumulative (default: difference/delta od)
  -q   char       quadrature method:  trapezoid with default 'T' for "B linear in tau" or 'E' for "B exponential in tau"
  -q   char       quadrature method:  default 'T'=trapez=linear in tau, 'E'=exponential in tau
  -T   float      surface temperature [K]  (default 0)
  -v              flag: verbose
  -x   Interval   lower,upper wavenumbers (comma separated pair of floats [no blanks!],
                                           default set according to range of optical depth in datafiles)
  -z   float      zenith angle (uplooking: 0dg, downlooking 180dg)
                  NOTE:  a horizontal view, i.e. angle=90dg, is not implemented!!!


  WARNING:
  od2ri  does not know the type of optical depth (delta or accumulated ...) given in the input file!
         ===> use the -C flag if you have (ac)cumulated optical depth
"""

_LICENSE_ = """\n
This file is part of the Py4CAtS package.

Authors:
Franz Schreier
DLR-IMF Oberpfaffenhofen
Copyright 2002 - 2019  The Py4CAtS authors

Py4CAtS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Py4CAtS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

####################################################################################################################################

try:   import numpy as np
except ImportError as msg:  raise SystemExit(str(msg) + '\nimport numeric python failed!')

from pairTypes import Interval
from cgsUnits import cgs
from moreFun import nexprl, cosdg
from misc import monotone
from oDepth import odArray, odRead, od_list2matrix, cod2dod, flipod, oDepth_zpT
from planck import Planck_Wavenumber
from radInt import riArray, riSave


####################################################################################################################################
####################################################################################################################################

def schwarzschild_BexpTau (vGrid, dodMatrix, temperature, zenithAngle=0.0, omega=0.0):
	""" Quadrature of Schwarzschild integral over optical depth.
	    (equivalent to Planck exponential in optical depth approximation)

	    For input and output arguments, see schwarzschid_BlinTau
	"""

	# NOTE: this function expects a nFreqs*nLayers matrix of delta optical depths,
	# i.e. all optical depths have already been interpolated to a common (the densest) wavenumber grid

	nLevels = dodMatrix.shape[1]+1  # nLayers+1

	# initialize (evtl. to be replaced by surface radiance as function input)
	radiance = 0.0
	# observation angle
	mue=abs(cosdg(zenithAngle));  secans = 1./mue

	coAlbedo = 1.0-omega  # 1-singleScatteringAlbedo

	if zenithAngle<90.:
		# resort for uplooking geometry:  optical depth starts with 0 at observer
		temperature = np.flipud(temperature); dodMatrix = np.fliplr(dodMatrix)
		print(' schwarzschild_BexpTau:  uplooking: flip levels upDown')

	bbLo = Planck_Wavenumber (vGrid, temperature[0])
	for l in range(1,nLevels):
		print('%2i %6.1fK  %s %10g %10g %8g' %
		      (l, temperature[l], 'od', min(dodMatrix[:,l-1]), max(dodMatrix[:,l-1]), np.mean(dodMatrix[:,l-1])), end=' ')
		if abs(temperature[l]-temperature[l-1])<0.1:
			dodMatrix[:,l] = dodMatrix[:,l-1] + dodMatrix[:,l]
			print('level', l, '  identical temp', temperature[l],temperature[l-1], '  merging!')
			continue  # with next level, i.e. skip rest of loop body
		bbHi       = Planck_Wavenumber (vGrid, temperature[l])
		beta       = np.log(bbLo/bbHi)/dodMatrix[:,l-1]
		transLayer = np.exp(-secans*coAlbedo*dodMatrix[:,l-1])
		radiance   = radiance*transLayer + coAlbedo*(bbLo*transLayer-bbHi)/(mue*beta-coAlbedo)
		print('-->  I  %10g %10g %8g' % (min(radiance), max(radiance), np.mean(radiance)))
		bbLo     = bbHi.copy()

	# sum optical depth over all layers
	transmission = np.exp(-dodMatrix.sum(1))

	return radiance, transmission


####################################################################################################################################

def schwarzschild_BlinTau (vGrid, dodMatrix, temperature, zenithAngle=0.0):
	""" Evaluate Schwarzschild integral equation:  integrate Planck * exp(-tau) d_tau  with tau=opticalDepth.

	    ARGUMENTS:
	    ----------
	    vGrid:         wavenumber (frequency) grid
	    dodMatrix:     delta/differential/layer optical depth "matrix"
	    temperature:   temperature at the levels (layer bounds)
	    zenithAngle:   viewing angle: 0dg=uplooking, 180dg=downloooking/nadir

	    RETURNS:
	    --------
	    radiance:      intensity in erg/s/(cm^2 sr cm-1) as a function of wavenumber
	    transmission:  total transmission as a function of wavenumber
	    """

	# NOTE: this function expects a nFreqs*nLayers matrix of delta optical depths,
	# i.e. all optical depths have already been interpolated to a common (the densest) wavenumber grid

	if zenithAngle>90.:  # nadir view - downlooking
		temperature = np.flipud(temperature)
		dodMatrix   = np.fliplr(dodMatrix)
		mue = -cosdg(zenithAngle)
	elif 0.0<=zenithAngle<90.:  # zenith view - uplooking
		mue = cosdg(zenithAngle)
	else:
		raise SystemExit ('ERROR --- schwarzschild_BlinTau:  observer zenith angle = 90dg (no "horizontal view")')

	if abs(mue-1.0)>0.001:
		dodMatrix = dodMatrix/mue

	# initialize
	bbLo = Planck_Wavenumber (vGrid, temperature[0])
	bbHi = Planck_Wavenumber (vGrid, temperature[-1])
	radiance    = bbLo - bbHi*np.exp(-dodMatrix.sum(1))
	sumOptDepth = np.zeros_like(vGrid)

	# loop over all layers
	for l in range(dodMatrix.shape[1]):
		bbHi         = Planck_Wavenumber (vGrid, temperature[l+1])
		deltaBB      = bbHi - bbLo
		radiance    += deltaBB * nexprl(dodMatrix[:,l]) * np.exp(-sumOptDepth)
		sumOptDepth += dodMatrix[:,l]
		bbLo         = bbHi.copy()

	# total transmission
	transmission = np.exp(-sumOptDepth)

	return radiance, transmission


####################################################################################################################################

def dod2ri (dodList, zenithAngle=0.0, tempBack=0.0, bExpTau=False):
	""" Evaluate Schwarzschild integral equation:  integrate Planck * exp(-tau) d_tau  with tau=opticalDepth.

 	    ARGUMENTS:
 	    ----------

 	    dodList:       delta (layer) optical depths (a list of odArray's)
 	    zenithAngle:   viewing zenith angle: 0dg=uplooking (default) ... 180dg=downloooking/nadir
                           NOTE:  a horizontal view with 90dg is not implemented!!!
 	    tempBack:      background temperature [K], default 0
 	    bExpTau:       flag: use B exponential in optical depth approximation, default B linear

 	    RETURNS:
 	    --------
 	    riArray:       a subclassed numpy array with radiance along with some attributes

	    NOTE:
	    -----
	    dod2ri  evaluates the Schwarzschild (and Beer) integral from the first to the last altitude,
	    i.e. lowest to highest for an uplooking observer (zenith<90dg) and highest to lowest for nadir view.
	    If you want radiance for an observer somewhere in between, give only the subset of relevant odArray's.
	    However, dod2ri does not consider an observer between the levels (layer bounds).
 	"""

 	# initial checks
	if not (isinstance(dodList,(list,tuple)) and all([isinstance(od,odArray) for od in dodList])):
		raise SystemExit ("ERROR --- dod2ri:  expected a list of odArray's, but ...")

	if 100.0<tempBack<=180.0 and zenithAngle>180.0:
		print("\nWARNING --- dod2ri:  did you mix up background temperature %f and zenith angle %f?!?" %
		      (tempBack, zenithAngle))

	# extract "attributes"
	zGrid, pGrid, temperature = oDepth_zpT(dodList)

	# viewing direction
	if monotone(zGrid)>0 and monotone(pGrid)<0:
		print('\n INFO --- dod2ri:  dodList (%i layers) sorted from BoA @ %.1f -> ToA @ %.1fkm' %
		      (len(dodList), cgs('!km',zGrid[0]), cgs('!km',zGrid[-1])))
	elif monotone(zGrid)<0 and monotone(pGrid)>0:
		print('\n INFO --- dod2ri:  dodList (%i layers) sorted from ToA @ %.1f -> BoA @ %.1fkm\n%s\n' %
		      (len(dodList), cgs('!km',zGrid[0]), cgs('!km',zGrid[-1]), ' flip up<-->down!'))
		dodList = flipod(dodList)
		zGrid   = np.flipud(zGrid)
		pGrid   = np.flipud(pGrid)
		temperature = np.flipud(temperature)
	else:
		raise SystemExit ("ERROR --- dod2ri:  both z and p grid monotone in or decreasing!")

	# interpolate all data to common, densest grid
	vGrid, dodMatrix = od_list2matrix (dodList)

	if bExpTau:
		radiance, transmission = schwarzschild_BexpTau (vGrid, dodMatrix, temperature, zenithAngle)
	else:
		radiance, transmission = schwarzschild_BlinTau (vGrid, dodMatrix, temperature, zenithAngle)

	if tempBack>0.0:
		bb = Planck_Wavenumber(vGrid,tempBack)
		# add background contribution to atmospheric signal
		radiance += bb*transmission
		print('%s %6.1fK %s %g %s %g' %
 		      (' Adding Planck BB background for', tempBack, '  ===>  ', bb.min(), '<= B <=', bb.max()))

	print(' dod2ri:     beer ===> %10.3g <= T <= %8f      schwarzschild ===> %10.3g <= I <= %10g      (%i points)' %
	      (min(transmission), max(transmission), min(radiance), max(radiance), len(vGrid)))

	return riArray (radiance, Interval(vGrid[0],vGrid[-1]),  tempBack, zenithAngle,
	                Interval(zGrid[0],zGrid[-1]), Interval(pGrid[0],pGrid[-1]), Interval(temperature.min(),temperature.max()))


####################################################################################################################################
####################################################################################################################################

def _od2ri_ (odFile, outFile=None, commentChar='#', cumOD=False,
             zenithAngle=180.,  zToA=0.0, zBoA=0.0, xLimits=None, omega=None,
             quad='T', tempBack=0.0, verbose=False):

	# read optical depth including some attributes, nb. temperature
	odList = odRead (odFile, zToA, zBoA, xLimits, commentChar, verbose)

	if cumOD:  # Subtract consecutive cumulated optical depths to delta (layer) optical depths
		odList = cod2dod(odList)

	if omega:  raise SystemExit('od2ri --- sorry, single scattering albedo not implemented')

	# assumes delta/layer optical depth as input
	bExpTau  = quad.upper()=='E'
	radiance = dod2ri (odList, zenithAngle, tempBack, bExpTau)

	# save radiance intensity
	riSave (radiance, outFile, commentChar='#')


####################################################################################################################################

if __name__ == "__main__":

	from command_parser import parse_command, standardOptions

	# parse the command, return (ideally) one file and some options
	opts = standardOptions + [  # h=help, c=commentChar, o=outFile
	       dict(ID='C', name='cumOD'),
               dict(ID='BoA', name='zBoA', type=float, constraint='zBoA>0.0'),
               dict(ID='ToA', name='zToA', type=float, constraint='zToA>0.0'),
	       dict(ID='x', name='xLimits', type=Interval, constraint='xLimits.lower>=0.0'),
               dict(ID='q', name='quad', type=str, default='T', constraint='quad.upper() in ["T", "E"]'),
               dict(ID='T', name='tempBack', type=float, default=0.0, constraint='tempBack>0.0'),
               dict(ID='w', name='omega', type=float, default=0.0, constraint='omega>=0.0'),
               dict(ID='z', name='zenithAngle', type=float, default=0., constraint='0<=zenithAngle<=180. and zenithAngle!=90.'),
               dict(ID='v', name='verbose')
               ]

	inFiles, options, commentChar, outFile = parse_command (opts,1)

	if 'h' in options:  raise SystemExit (__doc__ + "\n End of od2ri help")
	if 'about' in options:  raise SystemExit (_LICENSE_)

	# translate some options to boolean flags
	boolOptions = [opt.get('name',opt['ID']) for opt in opts if not ('type' in opt or opt['ID']=='h')]
	for key in boolOptions:                   options[key] = key in options

	_od2ri_ (inFiles[0], outFile, commentChar, **options)
